@@ -35,7 +35,7 @@ def make_bitwarden_container(api_key_flat_file:str = None):
     
     file_cred_obj = FlatFileCredContainer(
         file_path=api_key_flat_file,
-        allow_broad_permissions=False) # This ss very stubborn about reading a file that isn't locked down properly
+        allow_broad_permissions=False) # This is very stubborn about reading a file that isn't locked down properly
     file_contents = file_cred_obj.read()
     j = json.loads(file_contents)
 
@@ -58,9 +58,9 @@ class BitwardenCredContainer(CredContainerBase):
 
         Args:
             client_id (string): Username (email address)
-            client_secret (string): Password (Hashed, as would be treutnred by the has_password function)
+            client_secret (string): Password (Hashed, as would be returned by the has_password function)
             
-            session_key (string):  If passed, should correspond with a currently valid sesison key that corresponds with the '--session' 
+            session_key (string):  If passed, should correspond with a currently valid session key that corresponds with the '--session' 
                 for any command and/or the BW_SESSION environment variable.  Ultimately, this is the value we're after for any subsequent
                 interactions with the cli.  Thus, if supplied (and valid) this is really the only arg we need
         """
@@ -70,6 +70,12 @@ class BitwardenCredContainer(CredContainerBase):
         if which_bw is None:
             raise FileNotFoundError(f"This program wraps the bitwarden cli tool, 'bw', but it doesn't seem to be installed (or is not on PATH).  Please fix that and try again.  See:  https://bitwarden.com/help/article/cli/")
 
+        # We also need that bw needs to be at least a specific version
+        minimum_bw_version="1.18.1"
+        valid_version_installed = self._check_bw_version_is_valid(minimum_required_version=minimum_bw_version)
+        if valid_version_installed is False:
+            raise FileNotFoundError(f"The 'bw' command line is installed, but the version is too old.  Version {minimum_bw_version} or greater is required.  Please upgrade using your OS package manager.  Type 'bw --version' to check your version")
+
         # Pin client id and client secret to self
         self.client_id = client_id
         self.client_secret = client_secret
@@ -101,24 +107,72 @@ class BitwardenCredContainer(CredContainerBase):
             print("If you instantiated via a JSON config file, you can avoid this message in the future by adding the key 'email_address'")
 
         # Do the login flow.  This will ultimately pin the value for self.session_key if we didn't have a valid one already
-        if self.get_auth_status() != 'unlocked':
-            self.do_auth_and_unlock_flow()
+        if self._get_auth_status() != 'unlocked':
+            self._do_auth_and_unlock_flow()
 
         # At this point we should be unlocked for sure.  If not, we've failed miserably
-        if self.get_auth_status() != 'unlocked':
+        if self._get_auth_status() != 'unlocked':
             raise ValueError(f"The bitwarden vault should be unlocked now using the session key but it still isn't.  Something bad happened.  There might be a bug in this program.  Please troubleshoot.\nSession Key: {self.session_key}")
 
         # Load the vault and pin it to self.  Note that this will pin the vault with all passwords redacted
         self.vault_contents = None
-        self.load_vault()  # Sets self.vault_contents
+        self._load_vault()  # Sets self.vault_contents
 
-    def do_auth_and_unlock_flow(self):
+    def _check_bw_version_is_valid(self, minimum_required_version:str):
+        """
+        Checks the version of bitwarden.  We need 1.18.1 or higher to leverage reading password out of environment variables via --passwordenv
+        This method does not use the _do_bw_command() helper function because that passes the session key which we may not have yet
+
+        Args:
+            minimum_required_version (str): A minimum required version string.  Like "1.18.1"
+
+        Raises:
+            ValueError: If the 'bw --version' command results in an error for some reason
+
+        Returns:
+            [Boolean]: A flag telling us if the installed version is recent enough or not
+        """
+
+        # Get the bitwarden version from the command line. 
+        # We're purposely avoiding the _do_bw_command() method here.  
+        cmd = "bw --version"
+        result = subprocess.run(cmd, shell=True, capture_output=True)
+        return_code = result.returncode
+        std_out = result.stdout.decode('utf-8')
+        std_err = result.stderr.decode('utf-8')
+        if return_code != 0:
+            raise ValueError(f"The command '{cmd}' resulted in a non-zero exit code of {return_code}.\n{std_err}")
+        else:
+            bw_env_string = std_out.strip()
+
+        # Parse the installed version number
+        bw_version_parts = bw_env_string.split('.')
+        
+        major = int(bw_version_parts[0]) if len(bw_version_parts) >= 1 else 0
+        minor = int(bw_version_parts[1]) if len(bw_version_parts) >= 2 else 0
+        patch = int(bw_version_parts[2]) if len(bw_version_parts) >= 3 else 0
+
+        # Parse the necessary version number
+        required_bw_version_parts = minimum_required_version.split(".")
+        required_major = int(required_bw_version_parts[0]) if len(required_bw_version_parts) >= 1 else 0
+        required_minor = int(required_bw_version_parts[1]) if len(required_bw_version_parts) >= 2 else 0
+        required_patch = int(required_bw_version_parts[2]) if len(required_bw_version_parts) >= 3 else 0
+
+        # Is it up to snuff?
+        if major >= required_major and minor >= required_minor and patch >= required_patch:
+            valid_version = True
+        else:
+            valid_version = False
+        
+        return valid_version
+        
+    def _do_auth_and_unlock_flow(self):
         """
         Gently guides us through the necessary steps to get the vault into an unlocked state
         We need to go from 'unauthenticated' --> 'locked' --> 'unlocked'
         """
 
-        auth_status = self.get_auth_status()
+        auth_status = self._get_auth_status()
 
         # Bail out if we're already unlocked
         if auth_status == 'unlocked':
@@ -131,15 +185,15 @@ class BitwardenCredContainer(CredContainerBase):
         try:
             while auth_status != 'unlocked':
                 
-                auth_status = self.get_auth_status()
+                auth_status = self._get_auth_status()
                 
                 if auth_status == 'unauthenticated':
                     # Let's get authenticated
-                    self.log_in(password_environemt_variable=rand_variable_name)
+                    self._log_in(password_environemt_variable=rand_variable_name)
 
                 elif auth_status == 'locked':
                     # We are authenticated (That is, bitwarden is pointing to our account), but the vault is locked
-                    self.unlock(password_environemt_variable=rand_variable_name) # This method pins session_key to self
+                    self._unlock(password_environemt_variable=rand_variable_name) # This method pins session_key to self
                 elif auth_status == 'unlocked':
                     # We are authenticated and the vault is unlocked.  We can interact with it now
                     print("The vault is now unlocked.")
@@ -150,9 +204,9 @@ class BitwardenCredContainer(CredContainerBase):
             del os.environ[rand_variable_name] # Implicitly calls unsetenv
 
 
-    def log_in(self, password_environemt_variable:str):
+    def _log_in(self, password_environemt_variable:str):
         """
-        Walks is through the login process.  For deets, see 'bw login --help'
+        Walks is through the login process.  For details, see 'bw login --help'
 
         Args:
             password_environemt_variable (string): The name of an environment variable which contains our master password
@@ -164,9 +218,9 @@ class BitwardenCredContainer(CredContainerBase):
         # Now log in and point to the environment variable
         print ("Logging into Bitwarden...")
         cmd = f"bw login {self.email_address} --passwordenv {password_environemt_variable} --apikey {self.client_secret}"
-        result = subprocess.run(cmd, shell=True, capture_output=True)  #TODO:  Use helper function here
+        self._do_bw_command(command=cmd)
 
-    def unlock(self, password_environemt_variable:str):
+    def _unlock(self, password_environemt_variable:str):
         """
         Unlocks the vault after having previously logged in.  This action returns a session key
 
@@ -176,17 +230,10 @@ class BitwardenCredContainer(CredContainerBase):
         
         print ("Unlocking Bitwarden Vault...")
         cmd = f"bw unlock --passwordenv {password_environemt_variable} --raw"  #The raw flag simply prints the session key that we should use for subsequent requests
-        result = subprocess.run(cmd, shell=True, capture_output=True)  #TODO:  Use helper function here
+        session_key = self._do_bw_command(command=cmd)
+        self.session_key = session_key  # This can be set in the env var BW_SESSION or passed with a '--session' argument with any bw command    
 
-        if result.returncode == 0:
-            session_key = result.stdout.decode('utf-8')
-            self.session_key = session_key  #This can be set in the env var BW_SESSION or passed with a '--session' argument with any bw command    
-        else: 
-            err = result.stderr.decode('utf-8')
-            raise ValueError(err)
-
-
-    def get_bitwarden_status(self):
+    def _get_bitwarden_status(self):
         """
         Issues the 'bitwarden status' command, which returns a JSON object we can use to tell if we're logged in or not
         """
@@ -198,40 +245,46 @@ class BitwardenCredContainer(CredContainerBase):
             session_key_part = ""
 
         cmd = f"bw status{session_key_part}"
-        result = subprocess.run(cmd, shell=True, capture_output=True)  #TODO:  Use helper function here
-
-        if result.returncode != 0:
-            raise OSError(f"The command '{cmd}' resulted in return code of {result.returncode}:\n{result.stderr.decode('utf-8')}")
-        else:
-            ret_val = result.stdout.decode('utf-8')
+        ret_val = self._do_bw_command(command=cmd)
     
         return json.loads(ret_val)
 
-    def get_auth_status(self):
+    def _get_auth_status(self):
         """
         Returns the authentication status which according to 'bw status --help' should be one of these:
             "unauthenticated", "locked", "unlocked"
         """
 
-        return self.get_bitwarden_status()['status']
+        return self._get_bitwarden_status()['status']
         
 
-    def get_session_key(self):
+    def _retrieve_session_key(self):
         """
         Issues the command 'bw login --raw' which causes authentication to happen and returns a session key to be used for subsequent requests
         """
 
         # Get the status
-        self.get_bitwarden_status()
+        self._get_bitwarden_status()
 
         command = "bw login --raw"
-        result = subprocess.run(command, shell=True, capture_output=True)  #TODO:  Use helper function here
+        result = self._do_bw_command(command=command)
 
         print(f"Instantiated {type(self)} for username (email address) {self.username}")
         
 
-    def get_cred(self):
-        return super().get_cred()
+    def get_cred(self, guid:str):
+        """
+        A wrapper around the get_credentials_by_guid method.
+        Why?  Because this method is defined in the superclass and is intended to be overridden.
+        That's why.  Of course, it's perfectly fine to just call get_credentials_by_guid
+        Args:
+            guid (str): The Guid which we care to seek
+
+        Returns:
+            [dict]: Dict Containing the username and password in question
+        """
+
+        return self.get_credentials_by_guid(guid=guid) # Raises exceptions as needed
 
     def set_cred(self):
         return super().set_cred()
@@ -239,7 +292,7 @@ class BitwardenCredContainer(CredContainerBase):
     def delete_cred(self):
         return super().delete_cred()
 
-    def load_vault(self, force_reload=False):
+    def _load_vault(self, force_reload=False):
         """
         Gets the entire vault, removes all passwords and pins it to self for some client side interrogation
 
@@ -286,7 +339,7 @@ class BitwardenCredContainer(CredContainerBase):
         """
         Helper method.  Does a bitwarden cli command and passes the results back
         Args:
-            command (string):  The command to pass to the bw clie
+            command (string):  The command to pass to the bw cli
             raise_exceptions_on_non_zero (bool, optional): Controls exception raising if the command returns a non-zero code. Defaults to True.
         """
 
@@ -321,27 +374,87 @@ class BitwardenCredContainer(CredContainerBase):
             s = f"Object ID: {object_id}\tObject Type: {object_type}\tObject Name: {object_name}"
             
             if object_type != 'item':
-                raise ValueError(f"Encountered a non 'item' object type in the vault.  This is unexcpected. {s}")
+                raise ValueError(f"Encountered a non 'item' object type in the vault.  This is unexpected. {s}")
 
             print(s)
+
+    def get_vault_item_by_guid(self, guid:str):
+        """
+        Gets the item (JSON object) for a given GUID from the vault
+        been redacted previously
+
+        Args:
+            item_guid (str): [description]
+        """
+
+        # Load the vault that is pinned to self
+        if self.vault_contents is None:
+            self._load_vault()
+
+        # Try to find the item within the vault pinned to self.  This saves an unnecessary trip to to BW over the internet if it isn't there
+        sought_item = None
+        for item in self.vault_contents:
+            if item['id'] == guid:
+                sought_item = item
+                break
         
+        # Raise an exception if we didn't find the item
+        if sought_item is None:
+            raise ValueError(f"The item with GUID {guid} was not found in the vault.")
 
+        item_with_password = json.loads(self._do_bw_command(f"bw get item {guid}"))
 
+        return item_with_password
 
+    def get_credentials_by_guid(self, guid:str):
+        """
+        Returns the username and password (beneath the 'login' key) for a given item from the vault
+        This function simply wraps get_vault_item_by_guid, which will raise exceptions is the item is not in the vault
+        Args:
+            guid (str): The GUID for the item we want to retreive credentials for
+        """
+        
+        item_with_password = self.get_vault_item_by_guid(guid=guid)
+        login = item_with_password['login']
+        username = login['username']
+        password = login['password']
 
+        return dict(username=username, password=password)
 
-if __name__ == '__main__':
-    #TODO:  Really need to unit test this entire module thoroughly.  Write test cases
+    def get_username_by_guid(self, guid:str):
+        """
+        Returns the username for a given vault item by GUID.  Since we'll have the vault (without passwords) pinned to self already
+        We can just read that rather than pinging bitwarden again
 
+        Args:
+            guid (str): [description]
+        """
+        # Load the vault that is pinned to self
+        if self.vault_contents is None:
+            self._load_vault()
 
-    o = make_bitwarden_container()
-    o.print_items()
-    
+        # Try to find the item within the vault pinned to self.  This saves an unnecessary trip to to BW over the internet if it isn't there
+        sought_item = None
+        for item in self.vault_contents:
+            if item['id'] == guid:
+                sought_item = item
+                break
+        
+        # Raise an exception if we didn't find the item
+        if sought_item is None:
+            raise ValueError(f"The item with GUID {guid} was not found in the vault.")
 
+        return sought_item['login']['username']
 
+    def get_password_by_guid(self, guid:str):
+        """
+        Returns the password for a given item by GUID.  Wraps the get_credentials_by_guid method which in turn wraps get_vault_item_by_guid
+        (which raises exceptions if something is missing)
 
+        Args:
+            guid (str): [description]
+        """
 
-    # test = hash_password(password="p4ssw0rd", email="nobody@example.com")
-    # print(test)
+        creds = self.get_credentials_by_guid(guid=guid)
 
-    # print(prompt_for_credentials())
\ No newline at end of file
+        return creds['password']
